Module: BatchExperiment::UKP5Extractor

Extended by:
Extractor
Defined in:
lib/batch_experiment/sample_extractors.rb

Overview

Extractor for the output of the run_ukp5.out program available at github.com/henriquebecker91/masters. Not of interest for the majority of the users of this gem. Kept as example, and for this gem author personal use.

Class Method Summary collapse

Methods included from Extractor

extract, extract_from_lines, get_field, get_hfield, names

Class Method Details

.extract_from_lines(lines) ⇒ Object



32
33
34
35
36
# File 'lib/batch_experiment/sample_extractors.rb', line 32

def self.extract_from_lines(lines)
  ['Seconds', 'ext_time', 'ext_mem', 'opt'].map do | label |
    Extractor.get_field(lines, label)
  end
end

.namesObject



28
29
30
# File 'lib/batch_experiment/sample_extractors.rb', line 28

def self.names
  ['internal time', 'external time', 'external memory', 'opt']
end